@@ -1,15 +1,14 @@
 /* -*- mode:C; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
-/*
+/* 
  * This file is part of maemo-af-desktop
  *
- * Copyright (C) 2005,2006 Nokia Corporation.
+ * Copyright (C) 2005, 2006 Nokia Corporation.
  *
  * Contact: Karoliina Salminen <karoliina.t.salminen@nokia.com>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
- * as published by the Free Software Foundation; either version 2.1 of
- * the License, or (at your option) any later version.
+ * version 2.1 as published by the Free Software Foundation.
  *
  * This library is distributed in the hope that it will be useful, but
  * WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -23,13 +22,25 @@
  *
  */
 
-#include "hn-wm-watched-window.h"
-#include "osso-manager.h"
-
+#include <sys/types.h>
+#include <signal.h>
+#include <string.h>
 #include <X11/Xutil.h> 		/* For WMHints */
-
-/* Application relaunch indicator data*/
-#define RESTORED "restored"
+#include <X11/Xatom.h>
+#include <gdk/gdkx.h>
+#include <gtk/gtk.h> /* needed by hildon-navigator-main.h */
+#include <libosso.h>
+#include <log-functions.h>
+#include <hildon-widgets/hildon-defines.h>
+#include <hildon-widgets/hildon-banner.h>
+#include <hildon-widgets/hildon-note.h>
+#include "hildon-navigator-window.h"
+#include "hn-wm-watched-window.h"
+#include "hn-wm-watched-window-view.h"
+#include "hn-wm-watchable-app.h"
+#include "hn-entry-info.h"
+#include "hn-app-switcher.h"
+#include "osso-manager.h"
 
 #define PING_TIMEOUT_MESSAGE_STRING       _( "qgn_nc_apkil_notresponding" )
 #define PING_TIMEOUT_RESPONSE_STRING      _( "qgn_ib_apkil_responded" )
@@ -40,6 +51,40 @@
 
 #define HIBERNATION_TIMEMOUT 3000 /* as suggested by 31410#10 */
 
+extern HildonNavigatorWindow *tasknav;
+/*
+extern Navigator * task_nav;
+
+extern Navigator * task_nav;
+*/
+typedef char HNWMWatchedWindowFlags;
+
+typedef enum 
+{
+  HNWM_WIN_URGENT           = 1 << 0,
+  HNWM_WIN_NO_INITIAL_FOCUS = 1 << 1,
+
+  /*
+   *  Last dummy value for compile time check.
+   *  If you need values > HNWM_WIN_LAST, you have to change the
+   *  definition of HNWMWatchedWindowFlags to get more space.
+   */
+  HNWM_WIN_LAST           = 1 << 7
+}HNWMWatchedWindowFlagsEnum;
+
+
+/*
+ *  compile time assert making sure that the storage space for our flags is
+ *  big enough
+*/
+typedef char __window_compile_time_check_for_flags_size[
+     (guint)(1<<(sizeof(HNWMWatchedWindowFlags)*8))>(guint)HNWM_WIN_LAST ? 1:-1
+                                                       ];
+
+#define HNWM_WIN_SET_FLAG(a,f)    ((a)->flags |= (f))
+#define HNWM_WIN_UNSET_FLAG(a,f)  ((a)->flags &= ~(f))
+#define HNWM_WIN_IS_SET_FLAG(a,f) ((a)->flags & (f))
+
 struct HNWMWatchedWindow
 {
   Window                  xwin;
@@ -51,9 +96,10 @@ struct HNWMWatchedWindow
   HNWMWatchedWindowView  *view_active;
   GdkPixbuf              *pixb_icon;
   Window                  xwin_group;
-  gboolean                is_urgent;
-  gboolean                no_initial_focus;
   gchar                  *hibernation_key;
+  HNWMWatchedWindowFlags  flags;
+  HNEntryInfo            *info;
+  GdkWindow              *gdk_wrapper_win;
 };
 
 struct xwinv
@@ -62,8 +108,6 @@ struct xwinv
   gint    n_wins;
 };
 
-extern HNWM *hnwm;
-
 /* NOTE: for below caller traps x errors */
 static void
 hn_wm_watched_window_process_wm_state (HNWMWatchedWindow *win);
@@ -96,23 +140,23 @@ pixbuf_destroy (guchar *pixels, gpointer data)
   g_free(pixels);
 }
 
-
 static void
 hn_wm_watched_window_process_net_wm_icon (HNWMWatchedWindow *win)
 {
   gulong *data;
   gint    len = 0, offset, w, h, i;
   guchar *rgba_data, *p;
+  HNEntryInfo *info;
 
   rgba_data = p = NULL;
 
-  data
-    = hn_wm_util_get_win_prop_data_and_validate (hn_wm_watched_window_get_x_win (win),
-						 hnwm->atoms[HN_ATOM_NET_WM_ICON],
-						 XA_CARDINAL,
-						 0,
-						 0,
-						 &len);
+  data = hn_wm_util_get_win_prop_data_and_validate 
+                 (hn_wm_watched_window_get_x_win (win),
+                  hn_wm_get_atom(HN_ATOM_NET_WM_ICON),
+                  XA_CARDINAL,
+                  0,
+                  0,
+                  &len);
 
   HN_DBG("#### grabbing NET icon ####");
 
@@ -132,7 +176,7 @@ hn_wm_watched_window_process_net_wm_icon (HNWMWatchedWindow *win)
 
       HN_DBG("got w:%i, h:%im offset is %i\n", w, h, offset);
 
-      if (w == ICON_SIZE && h == ICON_SIZE)
+      if (w == 26 && h == 26)
 	break;
 
       offset += ((w*h) + 2);
@@ -179,9 +223,12 @@ hn_wm_watched_window_process_net_wm_icon (HNWMWatchedWindow *win)
     }
 
   /* FIXME: need to just update icon, also could be broke for views */
-  app_switcher_item_icon_sync (hnwm->app_switcher, win);
+  info = hn_wm_watched_window_peek_info (win);
 
- out:
+  if(info)
+    hn_app_switcher_changed (hn_wm_get_app_switcher(), info);
+
+out:
   if (data)
     XFree(data);
 }
@@ -198,17 +245,17 @@ hn_wm_watched_window_process_hildon_view_active (HNWMWatchedWindow *win)
     return;
 
   app = hn_wm_watched_window_get_app (win);
-
+  
   if (!app)
     return;
-
-  new_active_view_id
-    = hn_wm_util_get_win_prop_data_and_validate (hn_wm_watched_window_get_x_win (win),
-						 hnwm->atoms[HN_ATOM_HILDON_VIEW_ACTIVE],
-						 XA_WINDOW,
-						 32,
-						 0,
-						 NULL);
+  
+  new_active_view_id = hn_wm_util_get_win_prop_data_and_validate 
+                             (hn_wm_watched_window_get_x_win (win),
+                              hn_wm_get_atom(HN_ATOM_HILDON_VIEW_ACTIVE),
+                              XA_WINDOW,
+                              32,
+                              0,
+                              NULL);
 
   if (!new_active_view_id)
     return;
@@ -218,11 +265,12 @@ hn_wm_watched_window_process_hildon_view_active (HNWMWatchedWindow *win)
   /* Check the prop value is valid and not alreday active */
 
   if (current_active_view
-      && hn_wm_watched_window_view_get_id(current_active_view) == *new_active_view_id)
+      && hn_wm_watched_window_view_get_id (current_active_view) 
+            == *new_active_view_id)
     goto out;
 
   iter = hn_wm_watched_window_get_views (win);
-
+  
   /* Find what the view id matches for this window's views and
    * update.
    */
@@ -234,13 +282,13 @@ hn_wm_watched_window_process_hildon_view_active (HNWMWatchedWindow *win)
 
       if (hn_wm_watched_window_view_get_id (view) == *new_active_view_id)
 	{
-	  app_switcher_update_item (hnwm->app_switcher, win, view,
-				    AS_MENUITEM_TO_FIRST_POSITION);
+          HNEntryInfo *info;
 
+	  info = hn_wm_watched_window_view_get_info (view);
 	  hn_wm_watched_window_set_active_view (win, view);
+	  hn_app_switcher_changed_stack (hn_wm_get_app_switcher (), info);
 	  goto out;
 	}
-
       iter  = g_list_next(iter);
     }
 
@@ -264,14 +312,14 @@ hn_wm_watched_window_process_wm_name (HNWMWatchedWindow *win)
   win->name = NULL;
 
   /* Attempt to get UTF8 name */
-  win->name 
-    = hn_wm_util_get_win_prop_data_and_validate (win->xwin,
-                                                 hnwm->atoms[HN_ATOM_NET_WM_NAME],
-                                                 hnwm->atoms[HN_ATOM_UTF8_STRING],
-                                                 8,
-                                                 0,
-                                                 &n_items);
-
+  win->name = hn_wm_util_get_win_prop_data_and_validate 
+                    (win->xwin,
+                     hn_wm_get_atom(HN_ATOM_NET_WM_NAME),
+                     hn_wm_get_atom(HN_ATOM_UTF8_STRING),
+                     8,
+                     0,
+                     &n_items);
+  
   /* If that fails grab it basic way */
   if (win->name == NULL || n_items == 0 
       || !g_utf8_validate (win->name, n_items, NULL))
@@ -287,22 +335,22 @@ hn_wm_watched_window_process_wm_name (HNWMWatchedWindow *win)
 
   win->subname = NULL;
 
-  win->subname 
-    = hn_wm_util_get_win_prop_data_and_validate (win->xwin,
-                                                 hnwm->atoms[HN_ATOM_MB_WIN_SUB_NAME],
-                                                 XA_STRING,
-                                                 8,
-                                                 0,
-                                                 NULL);
-
+  win->subname = hn_wm_util_get_win_prop_data_and_validate 
+                      (win->xwin,
+                       hn_wm_get_atom(HN_ATOM_MB_WIN_SUB_NAME),
+                       XA_STRING,
+                       8,
+                       0,
+                       NULL);
+  
   view = hn_wm_watched_window_get_active_view(win);
-
+  
   /* Duplicate to topped view also */
   if (view)
     hn_wm_watched_window_view_set_name (view, win->name);
-
-  app_switcher_update_item (hnwm->app_switcher, win, view,
-			    AS_MENUITEM_SAME_POSITION);
+  
+  if(win->info)
+    hn_app_switcher_changed (hn_wm_get_app_switcher (), win->info);
 }
 
 static void
@@ -312,9 +360,9 @@ hn_wm_watched_window_process_wm_window_role (HNWMWatchedWindow *win)
 
   g_return_if_fail(win);
 
-  new_key =
-    hn_wm_compute_watched_window_hibernation_key(win->xwin,
-						 hn_wm_watched_window_get_app (win));
+  new_key = hn_wm_compute_watched_window_hibernation_key 
+                      (win->xwin,
+                       hn_wm_watched_window_get_app (win));
 
   if (new_key)
     {
@@ -339,24 +387,26 @@ hn_wm_watched_window_process_hibernation_prop (HNWMWatchedWindow *win)
   /* NOTE: foo has no 'value' if set the app is killable (hibernatable),
    *       deletes to unset
    */
-  foo = hn_wm_util_get_win_prop_data_and_validate (win->xwin,
-						   hnwm->atoms[HN_ATOM_HILDON_APP_KILLABLE],
-						   XA_STRING,
-						   8,
-						   0,
-						   NULL);
-
-  if(!foo)
-	{
-	  /*try the alias*/
-	  foo = hn_wm_util_get_win_prop_data_and_validate (win->xwin,
-						   hnwm->atoms[HN_ATOM_HILDON_ABLE_TO_HIBERNATE],
-						   XA_STRING,
-						   8,
-						   0,
-						   NULL);
-	}
+  foo = hn_wm_util_get_win_prop_data_and_validate 
+                     (win->xwin,
+                      hn_wm_get_atom(HN_ATOM_HILDON_APP_KILLABLE),
+                      XA_STRING,
+                      8,
+                      0,
+                      NULL);
 
+  if (!foo)
+    {
+      /*try the alias*/
+      foo = hn_wm_util_get_win_prop_data_and_validate 
+                    (win->xwin,
+                     hn_wm_get_atom(HN_ATOM_HILDON_ABLE_TO_HIBERNATE),
+                     XA_STRING,
+                     8,
+                     0,
+                     NULL);
+    }
+  
   hn_wm_watchable_app_set_able_to_hibernate (app,
 					     (foo != NULL) ? TRUE : FALSE );
   if (foo)
@@ -371,50 +421,26 @@ hn_wm_watched_window_process_wm_state (HNWMWatchedWindow *win)
 
   app = hn_wm_watched_window_get_app (win);
 
-  state
-    = hn_wm_util_get_win_prop_data_and_validate (win->xwin,
-						 hnwm->atoms[HN_ATOM_WM_STATE],
-						 hnwm->atoms[HN_ATOM_WM_STATE],
-						 0, /* FIXME: format */
-						 0,
-						 NULL);
-
+  state = hn_wm_util_get_win_prop_data_and_validate 
+                  (win->xwin,
+                   hn_wm_get_atom(HN_ATOM_WM_STATE),
+                   hn_wm_get_atom(HN_ATOM_WM_STATE),
+                   0, /* FIXME: format */
+                   0,
+                   NULL);
+  
   if (!state
       || (hn_wm_watchable_app_is_minimised(app) && state[0] == IconicState) )
     goto out;
-
+  
   if (state[0] == IconicState)
     {
       hn_wm_watchable_app_set_minimised (app, TRUE);
-
-      if (hn_wm_watched_window_get_views (win))
-	{
-	  GList      *iter;
-	  iter = hn_wm_watched_window_get_views (win);
-
-	  while (iter != NULL)
-	    {
-	      HNWMWatchedWindowView *view;
-
-	      view = (HNWMWatchedWindowView *)iter->data;
-
-	      app_switcher_update_item (hnwm->app_switcher, win, view,
-					AS_MENUITEM_TO_LAST_POSITION);
-
-	      iter  = g_list_next(iter);
-	    }
-	}
-      else
-	{
-	  /* Window with no views */
-	  app_switcher_update_item (hnwm->app_switcher, win, NULL,
-				    AS_MENUITEM_TO_LAST_POSITION );
-	}
     }
   else /* Assume non minimised state */
-   {
-     hn_wm_watchable_app_set_minimised (app, FALSE);
-   }
+    {
+      hn_wm_watchable_app_set_minimised (app, FALSE);
+    }
 
  out:
 
@@ -438,14 +464,27 @@ hn_wm_watched_window_process_wm_hints (HNWMWatchedWindow *win)
 
   win->xwin_group = wm_hints->window_group;
 
-  if (win->is_urgent != (wm_hints->flags & XUrgencyHint))
+  if (HNWM_WIN_IS_SET_FLAG(win,HNWM_WIN_URGENT) 
+                != (wm_hints->flags & XUrgencyHint))
     need_icon_sync = TRUE;
 
-  win->is_urgent = (wm_hints->flags & XUrgencyHint);
+  if(wm_hints->flags & XUrgencyHint)
+    {
+      HNWM_WIN_SET_FLAG(win,HNWM_WIN_URGENT);
+    }
+  else
+    {
+      HNWM_WIN_UNSET_FLAG(win,HNWM_WIN_URGENT);
+    }
 
   if (need_icon_sync)
-    app_switcher_item_icon_sync (hnwm->app_switcher, win);
+    {
+      HNEntryInfo *info = hn_wm_watched_window_peek_info (win);
 
+      if(info)
+        hn_app_switcher_changed (hn_wm_get_app_switcher (), info);
+    }
+  
   XFree(wm_hints);
 }
 
@@ -454,22 +493,24 @@ hn_wm_watched_window_process_net_wm_user_time (HNWMWatchedWindow *win)
 {
   gulong *data;
 
-  data
-    = hn_wm_util_get_win_prop_data_and_validate (hn_wm_watched_window_get_x_win (win),
-						 hnwm->atoms[HN_ATOM_NET_WM_USER_TIME],
-						 XA_CARDINAL,
-						 0,
-						 0,
-						 NULL);
-
+  data = hn_wm_util_get_win_prop_data_and_validate 
+                (hn_wm_watched_window_get_x_win (win),
+                 hn_wm_get_atom(HN_ATOM_NET_WM_USER_TIME),
+                 XA_CARDINAL,
+                 0,
+                 0,
+                 NULL);
+  
   HN_DBG("#### processing _NET_WM_USER_TIME ####");
 
   if (data == NULL)
     return;
 
   if (*data == 0)
-    win->no_initial_focus = TRUE;
-
+    {
+      HNWM_WIN_SET_FLAG(win,HNWM_WIN_NO_INITIAL_FOCUS);
+    }
+  
   if (data)
     XFree(data);
 }
@@ -480,17 +521,19 @@ hn_wm_watched_window_process_hildon_view_list (HNWMWatchedWindow *win)
   struct xwinv xwins;
   int          i;
   GList       *iter = NULL, *next_iter;
-
+  HNEntryInfo *info;
+  
   if (hn_wm_watched_window_is_hibernating(win))
     return;
 
-  xwins.wins
-    = hn_wm_util_get_win_prop_data_and_validate (win->xwin,
-						 hnwm->atoms[HN_ATOM_HILDON_VIEW_LIST],
-						 XA_WINDOW,
-						 32,
-						 0,
-						 &xwins.n_wins);
+  xwins.wins = hn_wm_util_get_win_prop_data_and_validate 
+                    (win->xwin,
+                     hn_wm_get_atom(HN_ATOM_HILDON_VIEW_LIST),
+                     XA_WINDOW,
+                     32,
+                     0,
+                     &xwins.n_wins);
+
   if (G_UNLIKELY(xwins.wins == NULL))
     return;
 
@@ -504,29 +547,29 @@ hn_wm_watched_window_process_hildon_view_list (HNWMWatchedWindow *win)
     {
       HNWMWatchedWindowView *view;
       gboolean               view_found;
-
+      
       view       = (HNWMWatchedWindowView *)iter->data;
       view_found = FALSE;
-
+      
       next_iter  = g_list_next(iter);
-
+      
       for (i=0; i < xwins.n_wins; i++)
 	if (xwins.wins[i] == hn_wm_watched_window_view_get_id (view))
 	  {
 	    view_found = TRUE;
 	    break;
 	  }
-
+      
       if (!view_found)
 	{
 	  /* view is not listed on client - delete the list entry */
 	  hn_wm_watched_window_remove_view  (win, view);
 	  hn_wm_watched_window_view_destroy (view);
 	}
-
+      
       iter = next_iter;
     }
-
+  
   /* Now add any new views in prop that we dont have listed */
 
   for (i=0; i < xwins.n_wins; i++)
@@ -551,34 +594,54 @@ hn_wm_watched_window_process_hildon_view_list (HNWMWatchedWindow *win)
 	}
 
       if (!view_found)
-	{
-	  HNWMWatchedWindowView *new_view;
+        {
+          HNWMWatchedWindowView *new_view;
 
-	  /* Not in internal list so its new, add it */
-	  new_view = hn_wm_watched_window_view_new (win, xwins.wins[i]);
+          /* Not in internal list so its new, add it */
+          new_view = hn_wm_watched_window_view_new (win, xwins.wins[i]);
+          
+          if (new_view)
+            hn_wm_watched_window_add_view (win, new_view);
 
-	  if (new_view)
-	    hn_wm_watched_window_add_view (win, new_view);
-	}
+          HN_DBG("adding view info to AS");
+          info = hn_wm_watched_window_view_get_info (new_view);
+          hn_app_switcher_add (hn_wm_get_app_switcher (), info);
+  
+          /* The window may have been 'viewless' before this 
+           * view was created to we need to remove the widget 
+           * ref for a viewless window  
+           */
+          if (hn_wm_watched_window_peek_info (win))
+            {
+              HN_DBG("adding first view; removing window info from AS");
+              
+              hn_app_switcher_remove(hn_wm_get_app_switcher (),
+                                     hn_wm_watched_window_peek_info (win));
+
+              /*
+               * since the window of multiviewed app does not figure in the AS,
+               * tell it to get rid of the info
+               */
+              hn_wm_watched_window_destroy_info (win);
+            }
+        }
     }
 
   if (xwins.wins)
     XFree(xwins.wins);
 }
 
-
 HNWMWatchedWindow*
 hn_wm_watched_window_new (Window            xid,
 			  HNWMWatchableApp *app)
 {
   HNWMWatchedWindow *win = NULL;
-  HNWMWatchedWindowView *active_view = NULL;
   gchar             *hkey;
   gboolean           win_found = FALSE;
-
   gpointer           win_ptr = NULL;
   gpointer           orig_key_ptr = NULL;
 
+
   /*  Check if this window is actually one coming out
    *  of 'hibernation', we use its WM_CLASS[+WM_WINDOW_ROLE] 
    *  to identify.
@@ -593,35 +656,40 @@ hn_wm_watched_window_new (Window            xid,
   
   g_return_val_if_fail(hkey, NULL);
 
-  win_found = g_hash_table_lookup_extended (hnwm->watched_windows_hibernating,
+  win_found = g_hash_table_lookup_extended (hn_wm_get_hibernating_windows(),
                                             (gconstpointer)hkey,
                                             & orig_key_ptr,
                                             & win_ptr);
-
+  
   HN_DBG("^^^^ win 0x%x ^^^^", (int)win);
 
   if (win_found)
     {
+      HNEntryInfo *info = NULL;
+
+      HN_DBG("New Window is from hibernation");
+
       win = (HNWMWatchedWindow*)win_ptr;
       
       /* Window already exists in our hibernating window hash.
        * There for we can reuse by just updating its var with
        * new window values
        */
-      g_hash_table_steal(hnwm->watched_windows_hibernating,
+      g_hash_table_steal(hn_wm_get_hibernating_windows(),
 			 hkey);
 
       /* We already have this value */
-      g_free(hkey);
+      g_free(hkey); 
       hkey = NULL;
 
       /* free the hash key */
       g_free(orig_key_ptr);
 
-      HN_DBG("window found in hibernation hash, updating icon");
+      /* mark the application as no longer hibernating */
+      hn_wm_watchable_app_set_hibernate (app, FALSE);
 
       /* As window has come out of hibernation and we still have all
-       * resources ( menu, views etc ) view creation etc wont fire
+       * resources ( views etc ) view creation etc wont fire
        * needed app_switcher updates. Therefore explicitly push
        * window up our selves.
        * Note, win->view_active will be NULL for viewless windows.
@@ -631,14 +699,12 @@ hn_wm_watched_window_new (Window            xid,
        * multi-view apps that just woken up.
        */
       if (win->views)
-        {
-          active_view = win->views->data;
-        }
-      
-      app_switcher_update_item (hnwm->app_switcher, 
-                                win, 
-                                active_view,
-                                AS_MENUITEM_TO_FIRST_POSITION);
+        info = hn_wm_watched_window_view_get_info (win->views->data);
+      else
+        info = hn_wm_watched_window_peek_info (win);
+
+      if (info)
+        hn_app_switcher_changed_stack (hn_wm_get_app_switcher(), info);
     }
   else
     win = g_new0 (HNWMWatchedWindow, 1);
@@ -682,13 +748,14 @@ hn_wm_watched_window_get_x_win (HNWMWatchedWindow *win)
 gboolean
 hn_wm_watched_window_is_urgent (HNWMWatchedWindow *win)
 {
-  return win->is_urgent;
+  return HNWM_WIN_IS_SET_FLAG(win,HNWM_WIN_URGENT);
+
 }
 
 gboolean
 hn_wm_watched_window_wants_no_initial_focus (HNWMWatchedWindow *win)
 {
-  return win->no_initial_focus;
+  return HNWM_WIN_IS_SET_FLAG(win,HNWM_WIN_NO_INITIAL_FOCUS);
 }
 
 const gchar*
@@ -711,12 +778,25 @@ hn_wm_watched_window_get_hibernation_key (HNWMWatchedWindow *win)
 
 void
 hn_wm_watched_window_set_name (HNWMWatchedWindow *win,
-			       gchar             *name)
+			       const gchar             *name)
 {
   if (win->name) g_free(win->name);
   win->name = g_strdup(name);
 }
 
+void
+hn_wm_watched_window_set_gdk_wrapper_win (HNWMWatchedWindow *win,
+                                          GdkWindow         *wrapper_win)
+{
+  if (win->gdk_wrapper_win) g_object_unref (win->gdk_wrapper_win);
+  win->gdk_wrapper_win = wrapper_win;
+}
+
+GdkWindow *
+hn_wm_watched_window_get_gdk_wrapper_win (HNWMWatchedWindow *win)
+{
+  return win->gdk_wrapper_win;
+}
 
 GtkWidget*
 hn_wm_watched_window_get_menu (HNWMWatchedWindow *win)
@@ -743,9 +823,19 @@ hn_wm_watched_window_set_menu (HNWMWatchedWindow *win,
 GList*
 hn_wm_watched_window_get_views (HNWMWatchedWindow *win)
 {
+  g_return_val_if_fail (win != NULL, NULL);
+  
   return win->views;
 }
 
+gint
+hn_wm_watched_window_get_n_views (HNWMWatchedWindow *win)
+{
+  g_return_val_if_fail (win != NULL, 0);
+  
+  return g_list_length (win->views);
+}
+
 void
 hn_wm_watched_window_add_view (HNWMWatchedWindow     *win,
 			       HNWMWatchedWindowView *view)
@@ -778,7 +868,7 @@ hn_wm_watched_window_get_active_view (HNWMWatchedWindow     *win)
   g_return_val_if_fail (win, NULL);
 
   if (win->view_active)
-    return win->view_active;
+  return win->view_active;
 
   /* We have no active view set atm so just return the first one. 
    * Works around some issues with hildon_app_new_with_view() not 
@@ -790,7 +880,6 @@ hn_wm_watched_window_get_active_view (HNWMWatchedWindow     *win)
   return NULL;
 }
 
-
 static gboolean
 hn_wm_watched_window_sigterm_timeout_cb (gpointer data)
 {
@@ -821,39 +910,50 @@ hn_wm_watched_window_sigterm_timeout_cb (gpointer data)
 }
 
 gboolean
-hn_wm_watched_window_attempt_signal_kill (HNWMWatchedWindow *win, int sig)
+hn_wm_watched_window_attempt_signal_kill (HNWMWatchedWindow *win,
+                                          int sig,
+                                          gboolean ensure)
 {
-  guchar *pid_result = NULL;
-  pid_t pid;
+  guint32 *pid_result = NULL;
 
-  pid_result = hn_wm_util_get_win_prop_data_and_validate (win->xwin,
-							  hnwm->atoms[HN_ATOM_NET_WM_PID],
-							  XA_CARDINAL,
-							  32,
-							  0,
-							  NULL);
+  pid_result = hn_wm_util_get_win_prop_data_and_validate 
+                     (win->xwin,
+                      hn_wm_get_atom(HN_ATOM_NET_WM_PID),
+                      XA_CARDINAL,
+                      32,
+                      0,
+                      NULL);
 
   if (pid_result == NULL)
     return FALSE;
 
-  pid = (pid_t)(pid_result[0]+256*pid_result[1]);
-
-  if(!pid)
+  if(!pid_result[0])
+    {
+      g_warning("PID was 0");
+      XFree(pid_result);
       return FALSE;
+    }
 
+  HN_DBG("Attempting to kill pid %d with signal %d",
+         pid_result[0], sig);
 
-  if (sig == SIGTERM)
+  if (sig == SIGTERM && ensure)
     {
       /* install timeout to check that the SIGTERM was handled */
       g_timeout_add (HIBERNATION_TIMEMOUT,
                      (GSourceFunc)hn_wm_watched_window_sigterm_timeout_cb,
-                     GINT_TO_POINTER (pid));
+                     GINT_TO_POINTER (pid_result[0]));
     }
   
-
-  if(kill(pid, sig /*SIGTERM*/) != 0)
+  /* we would like to use -1 to indicate that children should be
+   *  killed too, but it does not work for some reason
+   */
+  if(kill((pid_t)(pid_result[0]), sig) != 0)
     {
-      osso_log(LOG_ERR, "Failed to kill pid %d with SIGTERM", pid);
+      HN_DBG("... failed.");
+      
+      osso_log(LOG_ERR, "Failed to kill pid %d with signal %d",
+	       pid_result[0], sig);
 
       XFree(pid_result);
       return FALSE;
@@ -887,21 +987,20 @@ hn_wm_watched_window_awake (HNWMWatchedWindow *win)
   if (app)
     {
       /* Relaunch it with RESTORED */
+      hn_wm_watchable_app_set_launching (app, TRUE);
       osso_man = osso_manager_singleton_get_instance();
       osso_manager_launch(osso_man, hn_wm_watchable_app_get_service (app),
 			  RESTORED);
 
-      /* Remove it from our hibernating hash, on mapping it should
-       * get readded to regular WatchedWindows hash
-       */
-      hn_wm_watchable_app_set_hibernate (app, FALSE);
+      /* do not reset the hibernating flag yet -- we will do that when the app
+       * creates its first window
+      */
     }
 }
 
 void
 hn_wm_watched_window_destroy (HNWMWatchedWindow *win)
 {
-  GList                 *iter, *next_iter;
   HNWMWatchedWindowView *view;
   GtkWidget             *note;
 
@@ -913,28 +1012,45 @@ hn_wm_watched_window_destroy (HNWMWatchedWindow *win)
       HN_DBG("### Aborting destroying '%s' as hibernating ###", win->name);
       return;
     }
+  
+  /* If ping timeout note is displayed.. */
+  note = hn_wm_watchable_app_get_ping_timeout_note (win->app_parent);
 
-  if (win->menu_widget)
+  if (note)
     {
-      /* Only remove the widget if the app is *really* killed */
-      app_switcher_remove_item (hnwm->app_switcher, win->menu_widget);
+      /* Show infobanner */
+      gchar *response_message
+            = g_strdup_printf (PING_TIMEOUT_RESPONSE_STRING,
+                    win->name );
+  
+      /* Show the infoprint */
+      hildon_banner_show_information (NULL, NULL, response_message );
+
+      g_free (response_message);
+      
+      /* .. destroy the infonote */
+      gtk_widget_destroy( note );
+      hn_wm_watchable_app_set_ping_timeout_note (win->app_parent, NULL);
     }
-
-
-  /* Destroy the views too */
-  iter = win->views;
-
-  while (iter)
+  
+  if(win->info)
     {
-      next_iter = g_list_next(iter);
-
-      view = (HNWMWatchedWindowView *)iter->data;
+      /* only windows of multiwindow apps have their own info */
+      HN_DBG("a window of multiwindow application; removing info from AS");
+      hn_app_switcher_remove(hn_wm_get_app_switcher(),
+                             win->info);
+      hn_entry_info_free (win->info);
+      win->info = NULL;
+    }
+  
+  /* Destroy the views too */
+  while (win->views)
+    {
+      view = (HNWMWatchedWindowView *)win->views->data;
 
       hn_wm_watched_window_view_destroy (view);
 
-      win->views = g_list_delete_link(win->views, iter);
-
-      iter = next_iter;
+      win->views = g_list_delete_link(win->views, win->views);
     }
 
   if (win->name)
@@ -952,18 +1068,17 @@ hn_wm_watched_window_destroy (HNWMWatchedWindow *win)
   if(win->app_parent &&
      win == hn_wm_watchable_app_get_active_window(win->app_parent))
     {
-	  hn_wm_watchable_app_set_active_window(win->app_parent, NULL);
+      hn_wm_watchable_app_set_active_window(win->app_parent, NULL);
     }
 
-  /* If ping timeout note is displayed.. */
-  note = hn_wm_watchable_app_get_ping_timeout_note (win->app_parent);
+  if(hn_wm_get_active_window() == win)
+    hn_wm_reset_active_window();
 
-  if (note)
-    {
-	  /* .. destroy it */
-	  gtk_widget_destroy (note);
-      hn_wm_watchable_app_set_ping_timeout_note (win->app_parent, NULL);
-    }
+  if(hn_wm_get_last_active_window() == win)
+    hn_wm_reset_last_active_window();
+
+  if(win->gdk_wrapper_win)
+    g_object_unref (win->gdk_wrapper_win);
 
   g_free(win);
 }
@@ -1027,59 +1142,55 @@ void
 hn_wm_watched_window_reset_x_win (HNWMWatchedWindow * win)
 {
   g_return_if_fail (win);
-  
   win->xwin = None;
 }
 
-
 static void
 hn_wm_ping_timeout_dialog_response (GtkDialog *note, gint ret, gpointer data)
 {
   HNWMWatchedWindow *win = (HNWMWatchedWindow *)data;
   HNWMWatchableApp *app = hn_wm_watched_window_get_app (win);
-  
+
   gtk_widget_destroy (GTK_WIDGET(note));
   hn_wm_watchable_app_set_ping_timeout_note (app, NULL);
-  
-  if (ret == GTK_RESPONSE_OK) 
+
+  if (ret == GTK_RESPONSE_OK)
     {
       /* Kill the app */
-      if (!hn_wm_watched_window_attempt_signal_kill (win, SIGKILL)) 
+      if (!hn_wm_watched_window_attempt_signal_kill (win, SIGKILL, FALSE))
         {
           HN_DBG ("hn_wm_ping_timeout: "
                   "failed to kill application '%s'.",
                   win->name);
-        } 
+        }
     }
-
 }
 
-
-void 
+void
 hn_wm_ping_timeout (HNWMWatchedWindow *win)
 {
   GtkWidget *note;
 
   HNWMWatchableApp *app = hn_wm_watched_window_get_app (win);
-		
-  gchar *timeout_message 
+
+  gchar *timeout_message
     = g_strdup_printf (PING_TIMEOUT_MESSAGE_STRING, win->name );
-	
+
   /* FIXME: Do we need to check if the note already exists? */
   note = hn_wm_watchable_app_get_ping_timeout_note( app );
-	
-  if (note && GTK_IS_WIDGET(note)) 
+  
+  if (note && GTK_IS_WIDGET(note))
     {
       HN_DBG( "hn_wm_ping_timeout: "
               "the note already exists. ");
-    
+
       goto cleanup_and_exit;
     }
-	
+
   note = hildon_note_new_confirmation (NULL, timeout_message);
 
   hn_wm_watchable_app_set_ping_timeout_note (app, note);
-	
+
   hildon_note_set_button_texts (HILDON_NOTE(note),
                                 PING_TIMEOUT_BUTTON_OK_STRING,
                                 PING_TIMEOUT_BUTTON_CANCEL_STRING);
@@ -1089,31 +1200,179 @@ hn_wm_ping_timeout (HNWMWatchedWindow *win)
                     G_CALLBACK (hn_wm_ping_timeout_dialog_response),
                     win);
 
-  gtk_widget_show_all (note);  
-  
+  gtk_widget_show_all (note);
+
 cleanup_and_exit:
-  
+
   g_free( timeout_message );
 }
 
 
-void 
+void
 hn_wm_ping_timeout_cancel (HNWMWatchedWindow *win)
 {
   HNWMWatchableApp *app = hn_wm_watched_window_get_app( win );
-  
+
   GtkWidget *note = hn_wm_watchable_app_get_ping_timeout_note(app);
-  
-  gchar *response_message 
-    = g_strdup_printf (PING_TIMEOUT_RESPONSE_STRING, 
+
+  gchar *response_message
+    = g_strdup_printf (PING_TIMEOUT_RESPONSE_STRING,
                        win->name );
- 
+
   if (note && GTK_IS_WIDGET (note)) {
     gtk_dialog_response (GTK_DIALOG(note), GTK_RESPONSE_CANCEL);
   }
-  
+
   /* Show the infoprint */
   hildon_banner_show_information (NULL, NULL, response_message );
-  
+
   g_free (response_message);
 }
+
+
+/*
+ * Closes window and associated views (if any), handling hibernated
+ * applications according to the UI spec.
+*/
+void
+hn_wm_watched_window_close (HNWMWatchedWindow *win)
+{
+  XEvent ev;
+
+  g_return_if_fail(win);
+
+  if(!hn_wm_watched_window_is_hibernating(win))
+    {
+      /*
+       * To close the window, we dispatch _NET_CLOSE_WINDOW event to Matchbox.
+       * This will simulate the pressing of the close button on the app window,
+       * the app will do its thing, and in turn we will be notified about 
+       * changed client list, triggering the necessary updates of AS, etc.
+       */
+      memset(&ev, 0, sizeof(ev));
+
+      ev.xclient.type         = ClientMessage;
+      ev.xclient.window       = hn_wm_watched_window_get_x_win (win);
+      ev.xclient.message_type = hn_wm_get_atom(HN_ATOM_NET_CLOSE_WINDOW);
+      ev.xclient.format       = 32;
+      ev.xclient.data.l[0]    = CurrentTime;
+      ev.xclient.data.l[1]    = GDK_WINDOW_XID(gtk_widget_get_parent_window (GTK_WIDGET (tasknav)));
+  
+      gdk_error_trap_push();
+      XSendEvent(GDK_DISPLAY(), GDK_ROOT_WINDOW(), False,
+                 SubstructureRedirectMask, &ev);
+
+      XSync(GDK_DISPLAY(),FALSE);
+      gdk_error_trap_pop();
+    }
+#if 0
+  /*
+   *  Disabled util the infrastructure outside of TN for handling this is in
+   *  place
+   */
+  else if(/*have_dirty_data test comes here */)
+    {
+      /*
+       * If the window is hibernating and has dirty data, we need to remove the
+       * appropriate items from the AS, but we leave the window in our hash, to
+       * allow for the data to be recovered, and to comply with UI spec 1.0,
+       * paragraph 6.3
+      */
+      GList * views = hn_wm_watched_window_get_views (win);
+      
+      if(views)
+        {
+          while(views)
+            {
+              app_switcher_remove_item(hn_wm_get_app_switcher(),
+                                       hn_wm_watched_window_view_get_menu(
+                                       (HNWMWatchedWindowView*)views->data));
+
+              views = g_list_next(views);
+            }
+        }
+      else
+        {
+          app_switcher_remove_item(hn_wm_get_app_switcher(),
+                                   hn_wm_watched_window_get_menu (win));
+        }
+    }
+#endif
+  else /* hibernating window with no dirty data */
+    {
+      /* turn off the hibernation flag in our app to force full destruction */
+      HNWMWatchableApp * app = hn_wm_watched_window_get_app (win);
+
+      g_return_if_fail(app);
+
+      /* Set hibernate to FALSE and remove from hibernation hash as not
+       * hibernating anymore. Note g_hash_table_remove will call
+       * hn_wm_watched_window_destroy()
+      */
+      hn_wm_watchable_app_set_hibernate(app, FALSE);
+
+      g_hash_table_remove (hn_wm_get_hibernating_windows(), 
+                           hn_wm_watched_window_get_hibernation_key(win));
+
+      /* If the app has any windows left, we need to reset the hibernation 
+       * flag back to TRUE
+       */
+      if(hn_wm_watchable_app_has_hibernating_windows (app))
+        hn_wm_watchable_app_set_hibernate(app, TRUE);
+    }
+}
+
+void
+hn_wm_watched_window_set_info (HNWMWatchedWindow *win,
+			       HNEntryInfo       *info)
+{
+  if (win->info)
+    hn_entry_info_free (win->info);
+  
+  win->info = info;
+}
+
+HNEntryInfo *
+hn_wm_watched_window_peek_info (HNWMWatchedWindow *win)
+{
+  return win->info;
+}
+
+HNEntryInfo *
+hn_wm_watched_window_create_new_info (HNWMWatchedWindow *win)
+{
+  if(win->info)
+    {
+      /*
+       * refuse to creat a new info, in case the old one is already referenced
+       * by AS
+       */
+      g_warning("Window info alread exists");
+    }
+  else
+    {
+      win->info = hn_entry_info_new_from_window (win);
+    }
+  
+  return win->info;
+}
+
+void
+hn_wm_watched_window_destroy_info (HNWMWatchedWindow *win)
+{
+  if(win->info)
+    {
+      hn_entry_info_free (win->info);
+      win->info = NULL;
+    }
+}
+
+gboolean
+hn_wm_watched_window_is_active (HNWMWatchedWindow *win)
+{
+  if(win && win == hn_wm_get_active_window())
+    return TRUE;
+
+  return FALSE;
+}
+
