@@ -1,3 +1,26 @@
+/*
+ * This file is part of maemo-af-desktop
+ *
+ * Copyright (C) 2005, 2006 Nokia Corporation.
+ *
+ * Contact: Karoliina Salminen <karoliina.t.salminen@nokia.com>
+ *
+ * This library is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Lesser General Public License
+ * version 2.1 as published by the Free Software Foundation.
+ *
+ * This library is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
+ * Lesser General Public License for more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this library; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
+ * 02110-1301 USA
+ *
+ */
+
 /*
  * This is a VERY incomplete and even somewhat incorrect implementation of a
  * reader/parser for fdo's menu-spec compliant menus. Quite Maemo specific.
@@ -33,10 +56,12 @@
 
 typedef struct {
 	gchar *name;
+	gchar *comment;
 	gchar *icon;
 	gchar *exec;
 	gchar *service;
 	gchar *desktop_id;
+	gchar *text_domain;
 	gboolean allocated;
 } desktop_entry_t;
 
@@ -46,6 +71,7 @@ static GtkTreeIter *extras_iter = NULL;
 /* Function for composing an icon with the given emblem */
 static GdkPixbuf *add_emblem_to_icon( GdkPixbuf *icon, GdkPixbuf *emblem );
 
+
 /* Function for getting a list of all available files */
 GList *get_desktop_files(gchar *directory, GList *desktop_files);
 
@@ -68,9 +94,11 @@ static void destroy_desktop_item(gpointer data, gpointer null)
     desktop_entry_t * de = (desktop_entry_t *)data;
 
     g_free(de->name);
+    g_free(de->comment);
     g_free(de->icon);
     g_free(de->exec);
     g_free(de->service);
+    g_free(de->text_domain);
     g_free(de->desktop_id);
 
     g_free(de);
@@ -130,6 +158,66 @@ GdkPixbuf *get_icon(const char *icon_name, int icon_size)
 	return pixbuf;
 } /* static GdkPixbuf *get_icon() */
 
+/* If we can't find a matching icon size, use the one provided as parameter
+ * This is to avoid having two pixbufs in memory when not needed.
+ */
+GdkPixbuf *get_icon_with_fallback(const char *icon_name,
+                                  int icon_size,
+                                  GdkPixbuf *fallback)
+{
+    gint *icon_sizes;
+    gint i;
+    gboolean found;
+	GtkIconTheme *icon_theme;
+	GdkPixbuf *pixbuf = NULL;
+	GError *error = NULL;
+
+	if (icon_name) {
+		icon_theme = gtk_icon_theme_get_default();
+
+        icon_sizes = gtk_icon_theme_get_icon_sizes(icon_theme, icon_name);
+        
+        i = 0;
+        found = FALSE;
+        while (icon_sizes[i] != 0)
+        {
+            /* If found or has scalable icon (-1), use it
+             * Well, in theory at least. Maemo-Gtk+ doesn't return -1 at all
+             * it seems (scalable is a synonyme for "64" really)
+             */ 
+            if (icon_sizes[i] == icon_size || icon_sizes[i] == -1)
+            {
+                found = TRUE;
+                break;
+            }
+            i++;
+        }
+
+        g_free (icon_sizes);
+
+        if (!found)
+        {
+            g_object_ref(fallback);
+            return fallback;
+        }
+    
+		pixbuf = gtk_icon_theme_load_icon(icon_theme,
+				icon_name,
+				icon_size,
+				GTK_ICON_LOOKUP_NO_SVG, &error);
+
+		if (error) {
+			ULOG_ERR("Error loading icon '%s': %s\n", icon_name,
+					error->message);
+			g_error_free(error);
+			error = NULL;
+		}
+	} else {
+		ULOG_ERR("Error loading icon: no icon name\n");
+	}
+
+	return pixbuf;
+} /* static GdkPixbuf *get_icon_with_fallback() */
 
 GList *get_desktop_files(gchar *directory, GList *desktop_files)
 {
@@ -238,6 +326,13 @@ GList *get_desktop_files(gchar *directory, GList *desktop_files)
                             continue;
                         }
 
+ 			item->comment = g_key_file_get_locale_string(
+ 					key_file,
+ 					DESKTOP_ENTRY_GROUP,
+ 					DESKTOP_ENTRY_COMMENT_FIELD,
+ 					NULL,
+ 					NULL);
+
 			item->icon = g_key_file_get_string(
 					key_file,
 					DESKTOP_ENTRY_GROUP,
@@ -262,6 +357,12 @@ GList *get_desktop_files(gchar *directory, GList *desktop_files)
                 g_free(item->service);
                 item->service = s;
             }
+			
+            item->text_domain = g_key_file_get_string(
+					key_file,
+					DESKTOP_ENTRY_GROUP,
+					DESKTOP_ENTRY_TEXT_DOMAIN_FIELD,
+					NULL);
 
 			item->desktop_id = g_strdup( dir_entry->d_name );
 
@@ -325,12 +426,14 @@ static void read_menu_conf(const char *filename, GtkTreeStore *menu_tree,
 	gboolean doc_created = FALSE;
 
 	GdkPixbuf *app_icon                      = NULL;
+	GdkPixbuf *thumb_icon                    = NULL;
 
 	GdkPixbuf *favourite_icon                = NULL;
 	GdkPixbuf *favourite_open_icon           = NULL;
 	GdkPixbuf *favourite_closed_icon         = NULL;
 	
 	static GdkPixbuf *folder_icon            = NULL;
+	static GdkPixbuf *folder_thumb_icon      = NULL;
 	static GdkPixbuf *folder_open_icon       = NULL;
 	static GdkPixbuf *folder_closed_icon     = NULL;
 	
@@ -349,6 +452,9 @@ static void read_menu_conf(const char *filename, GtkTreeStore *menu_tree,
 	/* Get the folder icons */
 	if ( !folder_icon ) {
 		folder_icon = get_icon(ICON_FOLDER, ICON_SIZE);
+		folder_thumb_icon = get_icon_with_fallback(ICON_FOLDER,
+							ICON_THUMB_SIZE,
+							folder_icon);
 	}
 
 	if ( !folder_open_icon ) {
@@ -394,6 +500,8 @@ static void read_menu_conf(const char *filename, GtkTreeStore *menu_tree,
 				"",
 				TREE_MODEL_DESKTOP_ID,
 				"",
+                TREE_MODEL_COMMENT,
+                "",
 				-1);
 		
 		if ( favourite_icon ) {
@@ -481,6 +589,8 @@ static void read_menu_conf(const char *filename, GtkTreeStore *menu_tree,
 					key,
 					TREE_MODEL_ICON,
 					folder_icon,
+					TREE_MODEL_THUMB_ICON,
+					folder_thumb_icon,
 					TREE_MODEL_EMBLEM_EXPANDER_OPEN,
 					folder_open_icon,
 					TREE_MODEL_EMBLEM_EXPANDER_CLOSED,
@@ -491,6 +601,8 @@ static void read_menu_conf(const char *filename, GtkTreeStore *menu_tree,
 					"",
 					TREE_MODEL_DESKTOP_ID,
 					"",
+					TREE_MODEL_COMMENT,
+					"",
 					-1);
 
 			if ( strcmp( key, EXTRAS_MENU_STRING ) == 0 ) {
@@ -561,6 +673,18 @@ static void read_menu_conf(const char *filename, GtkTreeStore *menu_tree,
 						app_icon = get_icon( ICON_DEFAULT_APP,
 								ICON_SIZE );
 					}
+
+					/* Check if we have an thumb sized icon.. */
+					thumb_icon = get_icon_with_fallback( item->icon,
+					                                     ICON_THUMB_SIZE,
+					                                     app_icon );
+
+                    if ( !thumb_icon )
+                    {
+						/* .. or use the default */
+						thumb_icon = get_icon( ICON_DEFAULT_APP,
+								ICON_THUMB_SIZE );
+					}
 					
 					gtk_tree_store_set(menu_tree,
 							&child_iter,
@@ -568,6 +692,8 @@ static void read_menu_conf(const char *filename, GtkTreeStore *menu_tree,
 							item->name,
 							TREE_MODEL_ICON,
 							app_icon,
+							TREE_MODEL_THUMB_ICON,
+							thumb_icon,
 							TREE_MODEL_EMBLEM_EXPANDER_OPEN,
 							NULL,
 							TREE_MODEL_EMBLEM_EXPANDER_CLOSED,
@@ -578,12 +704,21 @@ static void read_menu_conf(const char *filename, GtkTreeStore *menu_tree,
 							item->service,
 							TREE_MODEL_DESKTOP_ID,
 							item->desktop_id,
+                            TREE_MODEL_TEXT_DOMAIN,
+                            item->text_domain,
+							TREE_MODEL_COMMENT,
+							item->comment,
 							-1);
 
 					if ( app_icon ){
 						g_object_unref( G_OBJECT( app_icon ) );
 						app_icon = NULL;
 					}
+					
+                    if ( thumb_icon ){
+						g_object_unref( G_OBJECT( thumb_icon ) );
+						thumb_icon = NULL;
+					}
 
 					/* Mark the item allocated */
 					item->allocated = TRUE;
@@ -611,12 +746,16 @@ static void read_menu_conf(const char *filename, GtkTreeStore *menu_tree,
 					NULL,
 					TREE_MODEL_EMBLEM_EXPANDER_CLOSED,
 					NULL,
+					TREE_MODEL_THUMB_ICON,
+					NULL,
 					TREE_MODEL_EXEC,
 					SEPARATOR_STRING,
 					TREE_MODEL_SERVICE,
 					SEPARATOR_STRING,
 					TREE_MODEL_DESKTOP_ID,
 					SEPARATOR_STRING,
+					TREE_MODEL_COMMENT,
+					SEPARATOR_STRING,
 					-1);
 		}
 
@@ -650,8 +789,8 @@ static void read_menu_conf(const char *filename, GtkTreeStore *menu_tree,
 
 			extras_iter = g_malloc0( sizeof( GtkTreeIter ) );
 
-			GtkTreePath *first_folder;
-			GtkTreePath *last_folder;
+			GtkTreePath *first_folder = NULL;
+			GtkTreePath *last_folder  = NULL;
 
 			find_first_and_last_root_level_folders(
 					GTK_TREE_MODEL( menu_tree ),
@@ -664,6 +803,7 @@ static void read_menu_conf(const char *filename, GtkTreeStore *menu_tree,
 						GTK_TREE_MODEL( menu_tree ),
 						&sibling, last_folder );
 
+                gtk_tree_path_free( last_folder );
 				gtk_tree_store_insert_after(
 						menu_tree, extras_iter,
 						NULL, &sibling );
@@ -673,6 +813,7 @@ static void read_menu_conf(const char *filename, GtkTreeStore *menu_tree,
 						GTK_TREE_MODEL( menu_tree ),
 						&sibling, first_folder );
 
+                gtk_tree_path_free( first_folder );
 				gtk_tree_store_insert_after(
 						menu_tree, extras_iter,
 						NULL, &sibling );
@@ -698,6 +839,8 @@ static void read_menu_conf(const char *filename, GtkTreeStore *menu_tree,
 					EXTRAS_MENU_STRING,
 					TREE_MODEL_ICON,
 					folder_icon,
+					TREE_MODEL_THUMB_ICON,
+					folder_thumb_icon,
 					TREE_MODEL_EMBLEM_EXPANDER_OPEN,
 					folder_open_icon,
 					TREE_MODEL_EMBLEM_EXPANDER_CLOSED,
@@ -732,6 +875,23 @@ static void read_menu_conf(const char *filename, GtkTreeStore *menu_tree,
 						&item_iter, extras_iter );
 
 				app_icon = get_icon( item->icon, ICON_SIZE );
+                if ( !app_icon ) {
+                  /* .. or use the default */
+                  app_icon = get_icon( ICON_DEFAULT_APP,
+                                       ICON_SIZE );
+                }
+
+                /* Check if we have an thumb sized icon.. */
+                thumb_icon = get_icon_with_fallback( item->icon,
+                                                     ICON_THUMB_SIZE,
+                                                     app_icon );
+
+                if ( !thumb_icon )
+                  {
+                    /* .. or use the default */
+                    thumb_icon = get_icon( ICON_DEFAULT_APP,
+                                           ICON_THUMB_SIZE );
+                  }
 
 				gtk_tree_store_set(menu_tree,
 						&item_iter,
@@ -739,6 +899,8 @@ static void read_menu_conf(const char *filename, GtkTreeStore *menu_tree,
 						item->name,
 						TREE_MODEL_ICON,
 						app_icon,
+						TREE_MODEL_THUMB_ICON,
+						thumb_icon,
 						TREE_MODEL_EMBLEM_EXPANDER_OPEN,
 						NULL,
 						TREE_MODEL_EMBLEM_EXPANDER_CLOSED,
@@ -747,6 +909,8 @@ static void read_menu_conf(const char *filename, GtkTreeStore *menu_tree,
 						item->exec,
 						TREE_MODEL_SERVICE,
 						item->service,
+                        TREE_MODEL_TEXT_DOMAIN,
+                        item->text_domain,
 						TREE_MODEL_DESKTOP_ID,
 						item->desktop_id,
 						-1 );
@@ -755,6 +919,11 @@ static void read_menu_conf(const char *filename, GtkTreeStore *menu_tree,
 					g_object_unref( G_OBJECT ( app_icon ) );
 					app_icon = NULL;
 				}
+				
+                if ( thumb_icon ) {
+					g_object_unref( G_OBJECT ( thumb_icon ) );
+					thumb_icon = NULL;
+				}
 
 				item->allocated = TRUE;
 			}
@@ -804,11 +973,14 @@ GtkTreeModel *get_menu_contents(void)
 			TREE_MODEL_COLUMNS,
 			G_TYPE_STRING,      /* Name */
 			GDK_TYPE_PIXBUF,   /* Icon */
+			GDK_TYPE_PIXBUF,   /* Thumb Icon */
 			GDK_TYPE_PIXBUF,   /* Icon & expander open emblem */
 			GDK_TYPE_PIXBUF,   /* Icon & expander closed emblem */
 			G_TYPE_STRING,	   /* Exec */
 			G_TYPE_STRING,	   /* Service  */
-			G_TYPE_STRING	   /* Desktop ID */
+			G_TYPE_STRING,	   /* Desktop ID */
+			G_TYPE_STRING,     /* Comment */
+			G_TYPE_STRING	   /* Text domain */
 			);
 		
 	gtk_tree_store_append(contents, &content_iter, NULL);
@@ -1033,6 +1205,7 @@ void find_first_and_last_root_level_folders( GtkTreeModel *model,
 
 	GtkTreeIter top_iter;
 	GtkTreeIter child_iter;
+	GtkTreeIter last_iter;
 
 	if ( !gtk_tree_model_get_iter_first( model, &top_iter ) ) {
 		ULOG_ERR( "find_first_and_last_root_level_folders: model is empty." );
@@ -1044,8 +1217,6 @@ void find_first_and_last_root_level_folders( GtkTreeModel *model,
 		return;
 	}
 
-	gchar *name = NULL;
-	gchar *desktop_id = NULL;
 
 	/* "Reset" the paths */
 	if ( first_folder != NULL ) {
@@ -1070,9 +1241,9 @@ void find_first_and_last_root_level_folders( GtkTreeModel *model,
 	 * if there's only one folder *last_folder will be  NULL 
 	 * after the loop finishes. */
 	do {
+        gchar *desktop_id = NULL;
+      
 		gtk_tree_model_get( model, &child_iter,
-				TREE_MODEL_NAME,
-				&name,
 				TREE_MODEL_DESKTOP_ID,
 				&desktop_id,
 				-1 );
@@ -1083,17 +1254,15 @@ void find_first_and_last_root_level_folders( GtkTreeModel *model,
 			if ( first_folder != NULL && *first_folder == NULL ) {
 				*first_folder = gtk_tree_model_get_path(
 						model, &child_iter );
-			} else if ( last_folder != NULL ) {
-				/* Free up the previous path of last folder */
-				if ( *last_folder != NULL ) {
-					gtk_tree_path_free( *last_folder );
-				}
-				/* Update the path of last folder */
-				*last_folder = gtk_tree_model_get_path(
-						model, &child_iter );
 			}
+           
+            last_iter = child_iter;
 		}
+
+        g_free( desktop_id );
 	} while ( gtk_tree_model_iter_next( model, &child_iter ) );
+
+    *last_folder = gtk_tree_model_get_path( model, &last_iter );
 }
 
 
@@ -1444,7 +1613,7 @@ gboolean set_menu_contents( GtkTreeModel *model )
 		return_value = TRUE;
 	}
 
-	fprintf(fp, "%s", (const char *) buffer->content);
+    fprintf(fp, "%s", (const char *) buffer->content);
 	fclose(fp);
 
 
@@ -1456,10 +1625,6 @@ gboolean set_menu_contents( GtkTreeModel *model )
 
 	xmlBufferFree( buffer );
 	xmlCleanupParser();
-	
-	if ( user_home_dir ) {
-		g_free( user_home_dir );
-	}
 
 	if ( user_menu_conf_file ) {
 		g_free( user_menu_conf_file );
